""""""

import os
import re
import sys
import readline
try:  # pragma: no cover
    from urllib import urlretrieve  # NOQA
except ImportError:  # pragma: no cover
    # PY3K
    from urllib.request import urlretrieve  # NOQA
import tempfile
from zipfile import ZipFile, is_zipfile
readline  # NOQA: make pyflakes happy, readline makes interactive mode keep history

import six
from importlib import import_module

from .rendering import render_structure
from .parsing import (
    parse_config,
    write_config,
    update_config,
    pretty_format_config,
)
from .bobexceptions import (
    ConfigurationError,
    TemplateConfigurationError,
    SkipQuestion,
    ValidationError,
)


DOTTED_REGEX = re.compile(r'^[a-zA-Z_.]+:[a-zA-Z_.]+$')


def resolve_dotted_path(name):
    module_name, dir_name = name.rsplit(':', 1)
    module = import_module(module_name)
    return os.path.join(os.path.dirname(module.__file__), dir_name)


def resolve_dotted_func(name):
    module_name, func_name = name.split(':')
    module = import_module(module_name)
    func = getattr(module, func_name, None)
    if func:
        return func
    else:
        raise ConfigurationError("There is no object named %s in module %s" % (func_name, module_name))


def maybe_resolve_dotted_func(name):
    if isinstance(name, six.string_types) and DOTTED_REGEX.match(name):
        return resolve_dotted_func(name)
    else:
        return name


def maybe_bool(value):
    if value == "True":
        return True
    if value == "False":
        return False
    else:
        return value


def parse_template(template_name):
    """Resolve template name into absolute path to the template
    and boolean if absolute path is temporary directory.
    """
    if template_name.startswith('http'):
        if '#' in template_name:
            url, subpath = template_name.rsplit('#', 1)
        else:
            url = template_name
            subpath = ''
        with tempfile.NamedTemporaryFile() as tmpfile:
            urlretrieve(url, tmpfile.name)
            if not is_zipfile(tmpfile.name):
                raise ConfigurationError("Not a zip file: %s" % tmpfile)
            zf = ZipFile(tmpfile)
            try:
                path = tempfile.mkdtemp()
                zf.extractall(path)
                return os.path.join(path, subpath), True
            finally:
                zf.close()

    if ':' in template_name:
        path = resolve_dotted_path(template_name)
    else:
        path = os.path.realpath(template_name)

    if not os.path.isdir(path):
        raise ConfigurationError('Template directory does not exist: %s' % path)
    return path, False


class Configurator(object):
    """Controller that figures out settings, asks questions and renders
    the directory structure.

    :param template: Template name
    :param target_directory: Filesystem path to a output directory
    :param bobconfig: Configuration for mr.bob behaviour
    :param variables: Given variables to questions
    :param defaults: Overriden defaults of the questions

    Additional to above settings, `Configurator` exposes following attributes:

    - :attr:`template_dir` is root directory of the template
    - :attr:`is_tempdir` if template directory is temporary (when using zipfile)
    - :attr:`templateconfig` dictionary parsed from `template` section
    - :attr:`questions` ordered list of `Question instances to be asked
    - :attr:`bobconfig` dictionary parsed from `mrbob` section of the config

    """

    def __init__(self,
                 template,
                 target_directory,
                 bobconfig=None,
                 variables=None,
                 defaults=None):
        if not bobconfig:
            bobconfig = {}
        if not variables:
            variables = {}
        if not defaults:
            defaults = {}
        self.variables = variables
        self.defaults = defaults
        self.target_directory = os.path.realpath(target_directory)

        # figure out template directory
        self.template_dir, self.is_tempdir = parse_template(template)

        # check if user is trying to specify output dir into template dir
        if self.template_dir in os.path.commonprefix([self.target_directory,
                                                      self.template_dir]):
            raise ConfigurationError('You can not use target directory inside the template')

        if not os.path.isdir(self.target_directory):
            os.makedirs(self.target_directory)

        # parse template configuration file
        template_config = os.path.join(self.template_dir, '.mrbob.ini')
        if not os.path.exists(template_config):
            raise TemplateConfigurationError('Config not found: %s' % template_config)
        self.config = parse_config(template_config)

        # parse questions from template configuration file
        self.raw_questions = self.config['questions']
        if self.raw_questions:
            self.questions = self.parse_questions(self.raw_questions, self.config['questions_order'])
        else:
            self.questions = []

        # parse bobconfig settings
        # TODO: move config resolution inside this function from cli.py
        self.bobconfig = update_config(bobconfig, self.config['mr.bob'])
        self.verbose = maybe_bool(self.bobconfig.get('verbose', False))
        self.quiet = maybe_bool(self.bobconfig.get('quiet', False))
        self.remember_answers = maybe_bool(self.bobconfig.get('remember_answers', False))
        self.ignored_files = self.bobconfig.get('ignored_files', '').split()
        self.ignored_directories = self.bobconfig.get('ignored_directories', '').split()

        # parse template settings
        self.templateconfig = self.config['template']
        self.post_render = [resolve_dotted_func(f) for f in self.templateconfig.get('post_render', '').split()]
        self.pre_render = [resolve_dotted_func(f) for f in self.templateconfig.get('pre_render', '').split()]
        self.post_ask = [resolve_dotted_func(f) for f in self.templateconfig.get('post_ask', '').split()]
        self.pre_ask = [resolve_dotted_func(f) for f in self.templateconfig.get('pre_ask', '').split()]
        self.renderer = resolve_dotted_func(
            self.templateconfig.get('renderer', 'mrbob.rendering:jinja2_renderer'))

    def render(self):
        """Render file structure given instance configuration. Basically calls
        :func:`mrbob.rendering.render_structure`.
        """
        if self.pre_render:
            for f in self.pre_render:
                f(self)
        render_structure(self.template_dir,
                         self.target_directory,
                         self.variables,
                         self.verbose,
                         self.renderer,
                         self.ignored_files,
                         self.ignored_directories)
        if self.remember_answers:
            write_config(os.path.join(self.target_directory, '.mrbob.ini'),
                         'variables',
                         self.variables)
        if self.post_render:
            for f in self.post_render:
                f(self)

    def parse_questions(self, config, order):
        q = []

        for question_key in order:
            key_parts = question_key.split('.')
            c = dict(config)
            for k in key_parts:
                c = c[k]
            # filter out subnamespaces
            c = dict([(k, v) for k, v in c.items() if not isinstance(v, dict)])
            question = Question(name=question_key, **c)
            q.append(question)
        return q

    def print_questions(self):  # pragma: no cover
        for line in pretty_format_config(self.raw_questions):
            print(line)
            # TODO: filter out lines without questions
            # TODO: seperate questions with a newline
            # TODO: keep order

    def ask_questions(self):
        """Loops through questions and asks for input if variable is not yet set.
        """
        if self.pre_ask:
            for f in self.pre_ask:
                f(self)
        # TODO: if users want to manipulate questions order, this is curently not possible.
        for question in self.questions:
            if question.name not in self.variables:
                self.variables[question.name] = question.ask(self)
        if self.post_ask:
            for f in self.post_ask:
                f(self)


class Question(object):
    """Question configuration. Parameters are used to configure questioning
    and possible validation of the answer.

    :param name: Unique, namespaced name of the question
    :param question: Question to be asked
    :param default: Default value of the question
    :param required: Is question required?
    :type required: bool
    :param command_prompt: Function to executed to ask the question given question text
    :param help: Optional help message
    :param pre_ask_question: Space limited functions in dotted notation to ask before the question is asked
    :param post_ask_question: Space limited functions in dotted notation to ask aster the question is asked
    :param **extra: Any extra parameters stored for possible extending of `Question` functionality

    Any of above parameters can be accessed as an attribute of `Question` instance.

    """

    def __init__(self,
                 name,
                 question,
                 default=None,
                 required=False,
                 command_prompt=six.moves.input,
                 pre_ask_question='',
                 post_ask_question='',
                 help="",
                 **extra):
        self.name = name
        self.question = question
        self.default = default
        self.required = maybe_bool(required)
        self.command_prompt = maybe_resolve_dotted_func(command_prompt)
        self.help = help
        self.pre_ask_question = [resolve_dotted_func(f) for f in pre_ask_question.split()]
        self.post_ask_question = [resolve_dotted_func(f) for f in post_ask_question.split()]
        self.extra = extra

    def __repr__(self):
        return six.u("<Question name=%(name)s question='%(question)s'"
                     " default=%(default)s required=%(required)s>") % self.__dict__

    def ask(self, configurator):
        """Eventually, ask the question.

        :param configurator: :class:`mrbob.configurator.Configurator` instance

        """
        correct_answer = None
        self.default = configurator.defaults.get(self.name, self.default)
        non_interactive = maybe_bool(configurator.bobconfig.get('non_interactive', False))
        if non_interactive:
            self.command_prompt = lambda x: ''

        try:
            while correct_answer is None:
                # hook: pre ask question
                for f in self.pre_ask_question:
                    try:
                        f(configurator, self)
                    except SkipQuestion:
                        return

                # prepare question
                if self.default:
                    question = six.u("--> %s [%s]: ") % (self.question, self.default)
                else:
                    question = six.u("--> %s: ") % self.question

                # ask question
                if six.PY3:  # pragma: no cover
                    answer = self.command_prompt(question).strip()
                else:  # pragma: no cover
                    answer = self.command_prompt(question.encode('utf-8')).strip().decode('utf-8')

                # display additional help
                if answer == "?":
                    if self.help:
                        print(self.help)
                    else:
                        print("There is no additional help text.")
                    continue

                if answer:
                    correct_answer = answer
                # if we don't have an answer, take default
                elif self.default is not None:
                    correct_answer = maybe_bool(self.default)
                # if we don't have an answer or default value and is required, reask
                elif self.required and not correct_answer:
                    if non_interactive:
                        raise ConfigurationError('non-interactive mode: question %s is required but not answered.' % self.name)
                    else:
                        # TODO: we don't cover this as coverage seems to ignore it
                        continue  # pragma: no cover
                else:
                    correct_answer = answer

                # hook: post ask question + validation
                for f in self.post_ask_question:
                    try:
                        correct_answer = f(configurator, self, correct_answer)
                    except ValidationError as e:
                        if non_interactive:
                            raise ConfigurationError('non-interactive mode: question %s failed validation.' % self.name)
                        else:
                            correct_answer = None
                            print("ERROR: " + str(e))
                            continue
        except KeyboardInterrupt:  # pragma: no cover
            print('\nExiting...')
            sys.exit(0)

        if not non_interactive:
            print('')
        return correct_answer



import unittest
import tempfile
import os
import shutil
from unittest import mock
import pkg_resources


class TestCLI(unittest.TestCase):
    def setUp(self):
        self.output_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.output_dir)

    def call_FUT(self, *args):
        from ..cli import main

        return main(args)

    def test_version(self):
        output = self.call_FUT("--version")
        version = pkg_resources.get_distribution("mr.bob").version
        self.assertEqual(output, version)

    def test_no_template_name(self):
        self.assertRaises(SystemExit, self.call_FUT)

    def test_no_template_directory(self):
        self.assertRaises(SystemExit, self.call_FUT, "foo")

    def test_dummy_template(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "empty")
        self.call_FUT("-O", self.output_dir, template_dir)

    @mock.patch("mrbob.cli.Configurator")
    def test_cleanup_tempdir(self, mock_Configurator):
        template_dir = tempfile.mkdtemp()
        mock_Configurator().is_tempdir.return_value = True
        mock_Configurator().template_dir = template_dir
        self.call_FUT(template_dir)
        self.assertFalse(os.path.exists(template_dir))

    def test_dummy_template_create_target_directory(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "empty")
        self.call_FUT("-O", os.path.join(self.output_dir, "notexist"), template_dir)
        self.assertTrue(os.path.isdir(self.output_dir))

    def test_list_questions(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "empty")
        self.call_FUT("--list-questions", template_dir)

    def test_missing_mrbobini_in_template(self):
        template_dir = os.path.join(
            os.path.dirname(__file__), "templates", "unbound", "etc"
        )
        self.assertRaises(
            SystemExit, self.call_FUT, "-O", self.output_dir, template_dir
        )

    def test_no_config_file(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates", "empty")
        self.assertRaises(SystemExit, self.call_FUT, "-c", "/notexists", template_dir)

    def test_zip_file_wrong(self):
        self.assertRaises(SystemExit, self.call_FUT, "foobar.zip")

    @mock.patch("mrbob.cli.os.path.expanduser")
    def test_configs_override_each_other(self, mock_expanduser):
        # global config
        globalconfig = tempfile.mkstemp()[1]
        mock_expanduser.return_value = globalconfig
        with open(globalconfig, "w") as f:
            f.write(
                """
[mr.bob]
only_global = glob
overriden_by_file = foo

[variables]
only_global = glob
overriden_by_file = file

[defaults]
only_global_2 = glob2
overriden_by_file_2 = file1

                    """
            )

        # config file
        tempconfig = tempfile.mkstemp()[1]
        with open(tempconfig, "w") as f:
            f.write(
                """
[mr.bob]
only_file = file
overriden_by_file = file1

[variables]
only_file = file
overriden_by_file = file1

[defaults]
only_file_2 = file1
overriden_by_file_2 = file2
                    """
            )

        template_dir = os.path.join(
            os.path.dirname(__file__), "templates", "multiconfig"
        )
        self.call_FUT("-v", "-n", "-O", self.output_dir, "-c", tempconfig, template_dir)
        with open(os.path.join(self.output_dir, "vars")) as f:
            output = f.read()
            self.assertEqual(output, "glob\nfile\nfile1\nglob2\nfile1\nfile2\n")

        # cleanup
        os.remove(tempconfig)



# -*- coding: utf-8 -*-

from shutil import rmtree
from tempfile import mkdtemp
import codecs
import os
import stat
import unittest

from unittest import mock
import six


class render_structureTest(unittest.TestCase):
    def setUp(self):
        import mrbob

        self.fs_tempdir = mkdtemp()
        self.fs_templates = os.path.abspath(
            os.path.join(os.path.dirname(mrbob.__file__), "tests", "templates")
        )

    def tearDown(self):
        rmtree(self.fs_tempdir)

    def call_FUT(
        self,
        template,
        variables,
        output_dir=None,
        verbose=True,
        renderer=None,
        ignored_files=[],
        ignored_directories=[],
    ):
        from ..rendering import render_structure
        from ..rendering import jinja2_renderer

        if output_dir is None:
            output_dir = self.fs_tempdir

        if renderer is None:
            renderer = jinja2_renderer

        render_structure(
            template,
            output_dir,
            variables,
            verbose,
            renderer,
            ignored_files,
            ignored_directories,
        )

    def test_subdirectories_created(self):
        from ..rendering import python_formatting_renderer

        self.call_FUT(
            os.path.join(self.fs_templates, "unbound"),
            dict(ip_addr="192.168.0.1", access_control="10.0.1.0/16 allow"),
            renderer=python_formatting_renderer,
        )
        self.assertTrue(os.path.exists("%s/%s" % (self.fs_tempdir, "/usr/local/etc")))

    def test_skip_mrbobini_copying(self):
        self.call_FUT(
            os.path.join(self.fs_templates, "skip_mrbobini"),
            dict(foo="123"),
        )
        self.assertTrue(os.path.exists("%s/%s" % (self.fs_tempdir, "test")))
        self.assertFalse(os.path.exists("%s/%s" % (self.fs_tempdir, ".mrbob.ini")))

    def test_ds_store(self):
        self.call_FUT(
            os.path.join(self.fs_templates, "ds_store"),
            dict(),
        )
        self.assertFalse(os.path.exists("%s/%s" % (self.fs_tempdir, ".mrbob.ini")))
        self.assertFalse(os.path.exists("%s/%s" % (self.fs_tempdir, ".DS_Store")))

    def test_ignored_files(self):
        self.call_FUT(
            os.path.join(self.fs_templates, "ignored"),
            dict(),
            ignored_files=["ignored", "*.txt", ".mrbob.ini"],
        )
        self.assertFalse(os.path.exists("%s/%s" % (self.fs_tempdir, ".mrbob.ini")))
        self.assertFalse(os.path.exists("%s/%s" % (self.fs_tempdir, "ignored")))
        self.assertFalse(os.path.exists("%s/%s" % (self.fs_tempdir, "ignored.txt")))
        self.assertTrue(os.path.exists("%s/%s" % (self.fs_tempdir, "not_ignored")))

    def test_ignored_directories(self):
        self.call_FUT(
            os.path.join(self.fs_templates, "ignored_dirs"),
            dict(),
            ignored_directories=["ignored", "*_stuff"],
        )
        self.assertFalse(os.path.exists("%s/%s" % (self.fs_tempdir, "ignored")))
        self.assertFalse(os.path.exists("%s/%s" % (self.fs_tempdir, "ignored_stuff")))
        self.assertTrue(os.path.exists("%s/%s" % (self.fs_tempdir, "not_ignored")))

    def test_encoding_is_utf8(self):
        from ..rendering import python_formatting_renderer

        if six.PY3:  # pragma: no cover
            folder_name = "encodingč"
        else:  # pragma: no cover
            folder_name = "encodingč".decode("utf-8")

        self.call_FUT(
            os.path.join(self.fs_templates, folder_name),
            dict(),
            renderer=python_formatting_renderer,
        )

        if six.PY3:  # pragma: no cover
            file_name = "mapča/ća"
            expected = "Ćača.\n"
        else:  # pragma: no cover
            file_name = "mapča/ća".decode("utf-8")
            expected = "Ćača.\n".decode("utf-8")

        with codecs.open(os.path.join(self.fs_tempdir, file_name), "r", "utf-8") as f:
            self.assertEqual(f.read(), expected)

    def test_string_replacement(self):
        from ..rendering import python_formatting_renderer

        self.call_FUT(
            os.path.join(self.fs_templates, "unbound"),
            dict(ip_addr="192.168.0.1", access_control="10.0.1.0/16 allow"),
            verbose=False,
            renderer=python_formatting_renderer,
        )
        fs_unbound_conf = os.path.join(
            self.fs_tempdir, "usr/local/etc/unbound/unbound.conf"
        )
        self.assertTrue("interface: 192.168.0.1" in open(fs_unbound_conf).read())

    def test_directory_is_renamed(self):
        from ..rendering import python_formatting_renderer

        self.call_FUT(
            os.path.join(self.fs_templates, "renamedir"),
            dict(name="blubber"),
            verbose=False,
            renderer=python_formatting_renderer,
        )
        self.assertTrue(
            os.path.exists("%s/%s" % (self.fs_tempdir, "/partsblubber/part"))
        )

    def test_copied_file_is_renamed(self):
        from ..rendering import python_formatting_renderer

        self.call_FUT(
            os.path.join(self.fs_templates, "renamedfile"),
            dict(name="blubber"),
            verbose=False,
            renderer=python_formatting_renderer,
        )
        self.assertTrue(os.path.exists("%s/%s" % (self.fs_tempdir, "/foo.blubber.rst")))

    def test_rendered_file_is_renamed(self):
        from ..rendering import python_formatting_renderer

        self.call_FUT(
            os.path.join(self.fs_templates, "renamedtemplate"),
            dict(name="blubber", module="blather"),
            verbose=False,
            renderer=python_formatting_renderer,
        )
        fs_rendered = "%s/%s" % (self.fs_tempdir, "/blubber_endpoint.py")
        self.assertTrue(os.path.exists(fs_rendered))
        self.assertTrue("from blather import bar" in open(fs_rendered).read())

    def test_rendered_file_is_renamed_dotted_name(self):
        from ..rendering import python_formatting_renderer

        self.call_FUT(
            os.path.join(self.fs_templates, "renamedtemplate2"),
            {"author.name": "foo"},
            verbose=False,
            renderer=python_formatting_renderer,
        )
        self.assertTrue(os.path.exists("%s/%s" % (self.fs_tempdir, "/foo_endpoint.py")))

    def test_compount_renaming(self):
        """all of the above edgecases in one fixture"""
        from ..rendering import python_formatting_renderer

        self.call_FUT(
            os.path.join(self.fs_templates, "renamed"),
            dict(name="blubber", module="blather"),
            verbose=False,
            renderer=python_formatting_renderer,
        )
        fs_rendered = "%s/%s" % (
            self.fs_tempdir,
            "/blatherparts/blubber_etc/blubber.conf",
        )
        self.assertTrue(os.path.exists(fs_rendered))
        self.assertTrue("blather = blubber" in open(fs_rendered).read())


class render_templateTest(unittest.TestCase):
    def setUp(self):
        import mrbob

        self.fs_tempdir = mkdtemp()
        self.fs_templates = os.path.abspath(
            os.path.join(os.path.dirname(mrbob.__file__), "tests", "templates")
        )

    def tearDown(self):
        rmtree(self.fs_tempdir)

    def call_FUT(
        self, template, variables, output_dir=None, verbose=False, renderer=None
    ):
        from ..rendering import render_template
        from ..rendering import python_formatting_renderer

        if output_dir is None:
            output_dir = self.fs_tempdir

        if renderer is None:
            renderer = python_formatting_renderer

        return render_template(
            template,
            output_dir,
            variables,
            verbose,
            renderer,
        )

    def test_render_copy(self):
        """if the source is not a template, it is copied."""
        fs_source = os.path.join(self.fs_templates, "unbound/etc/rc.conf")

        fs_rendered = self.call_FUT(
            fs_source, dict(ip_addr="192.168.0.1", access_control="10.0.1.0/16 allow")
        )
        self.assertTrue(fs_rendered.endswith("rc.conf"))
        with open(fs_source) as f1:
            with open(fs_rendered) as f2:
                self.assertEqual(f1.read(), f2.read())

    def test_render_template(self):
        """if the source is a template, it is rendered and the target file drops
        the `.bob` suffix."""
        fs_source = os.path.join(
            self.fs_templates, "unbound/usr/local/etc/unbound/unbound.conf.bob"
        )
        fs_rendered = self.call_FUT(
            fs_source, dict(ip_addr="192.168.0.1", access_control="10.0.1.0/16 allow")
        )
        self.assertTrue(fs_rendered.endswith("/unbound.conf"))
        self.assertTrue("interface: 192.168.0.1" in open(fs_rendered).read())

    def test_rendered_permissions_preserved(self):
        fs_source = os.path.join(
            self.fs_templates, "unbound/usr/local/etc/unbound/unbound.conf.bob"
        )
        os.chmod(fs_source, 771)
        fs_rendered = self.call_FUT(
            fs_source, dict(ip_addr="192.168.0.1", access_control="10.0.1.0/16 allow")
        )
        self.assertEqual(stat.S_IMODE(os.stat(fs_rendered).st_mode), 771)

    def test_render_missing_key(self):
        t = os.path.join(
            self.fs_templates, "unbound/usr/local/etc/unbound/unbound.conf.bob"
        )

        self.assertRaises(KeyError, self.call_FUT, t, dict())

    def test_render_namespace(self):
        t = os.path.join(self.fs_templates, "missing_namespace_key/foo.bob")

        filename = self.call_FUT(t, {"foo.bar": "1"})
        with open(filename) as f:
            self.assertEqual(f.read(), "1\n")

    def test_render_namespace_jinja2(self):
        from ..rendering import jinja2_renderer

        t = os.path.join(self.fs_templates, "missing_namespace_key/foo_jinja2.bob")

        filename = self.call_FUT(t, {"foo.bar": "2"}, renderer=jinja2_renderer)
        with open(filename) as f:
            self.assertEqual(f.read(), "2\n")

    def test_render_newline(self):
        from ..rendering import jinja2_renderer

        t = os.path.join(self.fs_templates, "missing_namespace_key/foo_jinja2.bob")

        tfile = open(t, "r")
        self.assertEqual(tfile.read(), "{{{foo.bar}}}\n")

        filename = self.call_FUT(t, {"foo.bar": "2"}, renderer=jinja2_renderer)
        with open(filename) as f:
            self.assertEqual(f.read(), "2\n")

    def test_render_namespace_missing_key(self):
        t = os.path.join(self.fs_templates, "missing_namespace_key/foo.bob")

        self.assertRaises(KeyError, self.call_FUT, t, {})

    def test_render_namespace_missing_key_jinja2(self):
        from jinja2 import UndefinedError
        from ..rendering import jinja2_renderer

        t = os.path.join(self.fs_templates, "missing_namespace_key/foo_jinja2.bob")

        self.assertRaises(
            UndefinedError, self.call_FUT, t, {}, renderer=jinja2_renderer
        )

    def test_jinja2_strict_undefined(self):
        from jinja2 import UndefinedError
        from ..rendering import jinja2_renderer

        t = os.path.join(self.fs_templates, "strict_undefined.bob")

        self.assertRaises(
            UndefinedError, self.call_FUT, t, {}, renderer=jinja2_renderer
        )


class render_filenameTest(unittest.TestCase):
    def call_FUT(self, filename, variables):
        from ..rendering import render_filename

        return render_filename(filename, variables)

    def test_filename_substitution(self):
        t = self.call_FUT("em0_+ip_addr+.conf", dict(ip_addr="127.0.0.1"))
        self.assertEqual(t, "em0_127.0.0.1.conf")

    def test_filename_nested(self):
        t = self.call_FUT("em0_+ip.addr+.conf", {"ip.addr": "127.0.0.1"})
        self.assertEqual(t, "em0_127.0.0.1.conf")

    def test_multiple_filename_substitution(self):
        t = self.call_FUT(
            "+device+_+ip_addr+.conf", dict(ip_addr="127.0.0.1", device="em0")
        )
        self.assertEqual(t, "em0_127.0.0.1.conf")

    def test_single_plus_not_substituted(self):
        t = self.call_FUT("foo+bar", dict(foo="127.0.0.1", bar="em0"))
        self.assertEqual(t, "foo+bar")

    def test_no_substitution(self):
        t = self.call_FUT("foobar", dict(foo="127.0.0.1", bar="em0"))
        self.assertEqual(t, "foobar")

    def test_pluses_in_path(self):
        t = self.call_FUT("+/bla/+/+bar+", dict(bar="em0"))
        self.assertEqual(t, "+/bla/+/em0")

    @mock.patch("mrbob.rendering.os", sep="\\")
    def test_pluses_in_pathwindows(self, mock_sep):
        t = self.call_FUT("+\\bla\\+\\+bar+", dict(bar="em0"))
        self.assertEqual(t, "+\\bla\\+\\em0")

    def test_missing_key(self):
        self.assertRaises(KeyError, self.call_FUT, "foo+bar+blub", dict())


class parse_variablesTest(unittest.TestCase):
    def test_complex_example(self):
        from ..rendering import parse_variables

        variables = {
            "author.name": "foobar",
            "author.age": "23",
            "license": "BSD",
            "foo.bar.zar.mar": "foo",
        }
        vars_ = parse_variables(variables)
        self.assertEqual(
            set(vars_.keys()),
            set(["foo", "license", "author"]),
        )
        self.assertEqual(
            set(vars_["author"].items()),
            set([("name", "foobar"), ("age", "23")]),
        )
        self.assertEqual(
            set(vars_["foo"]["bar"]["zar"].items()),
            set([("mar", "foo")]),
        )

        # there is no such key in this namespace
        self.assertRaises(KeyError, lambda x: vars_["author"][x], "foo")



# -*- coding: utf-8 -*-
import os
import unittest
import tempfile
import codecs
from collections import OrderedDict  # NOQA

import six

from unittest import mock


class parse_configTest(unittest.TestCase):

    def call_FUT(self, configname='example.ini'):
        import mrbob
        from ..parsing import parse_config

        if not configname.startswith('http'):
            f = os.path.abspath(
                os.path.join(
                    os.path.dirname(mrbob.__file__), 'tests', configname)
            )
        else:
            f = configname
        return parse_config(f)

    def test_parse_variable(self):
        c = self.call_FUT()
        self.assertEqual(c['variables']['name'], 'Bob')

    def test_parse_nested_variable(self):
        c = self.call_FUT()
        self.assertEqual(c['variables']['host.ip_addr'], '10.0.10.120')

    def test_parse_2nd_level_nested_variable(self):
        c = self.call_FUT()
        self.assertEqual(c['variables']['webserver.foo.bar'], 'barf')

    def test_parse_nested_variable_out_of_order(self):
        c = self.call_FUT('example2.ini')
        self.assertEqual(c['variables']['webserver.foo.bar'], 'barf2')
        self.assertEqual(c['variables']['webserver.ip_addr'], '127.0.0.3')

    def test_ignored_files(self):
        c = self.call_FUT('ignored.ini')
        self.assertTrue('ignored' in c['mr.bob']['ignored_files'])
        self.assertTrue('*.txt' in c['mr.bob']['ignored_files'])

    def test_ignored_directories(self):
        c = self.call_FUT('ignored_dirs.ini')
        self.assertTrue('ignored' in c['mr.bob']['ignored_directories'])
        self.assertTrue('*_stuff' in c['mr.bob']['ignored_directories'])

    def test_parse_deeply_nested_variables(self):
        c = self.call_FUT('example5.ini')
        expected_config = {
            'mr.bob': {},
            'variables': {'a.b.c.d': 'foo', 'a.b.c.f': 'bar', 'name': 'Bob'},
            'questions': {'a': {'b': {'c': {'d': 'foo', 'f': 'bar'}}}, 'name': 'Bob'},
            'template': {},
            'defaults': {},
            'questions_order': [],
        }
        self.assertEqual(c, expected_config)

    def test_parse_config_utf8(self):
        from ..parsing import pretty_format_config
        c = self.call_FUT('example6.ini')
        output_variables = pretty_format_config(c['variables'])
        output_questions = pretty_format_config(c['questions'])
        if six.PY3:  # pragma: no cover
            expected_output = [
                'name = Čebula',
            ]
        else:  # pragma: no cover
            expected_output = [
                'name = Čebula'.decode('utf-8'),
            ]

        self.assertEqual(output_variables, expected_output)
        self.assertEqual(output_questions, expected_output)

    def test_parse_config(self):
        from ..parsing import pretty_format_config
        c = self.call_FUT()
        output = pretty_format_config(c['variables'])
        expected_output = [
            'host.ip_addr = 10.0.10.120',
            'name = Bob',
            'webserver.foo.bar = barf',
            'webserver.fqdn = mrbob.10.0.10.120.xip.io',
            'webserver.ip_addr = 127.0.0.2',
        ]
        self.assertEqual(output, expected_output)

    def test_question_order(self):
        c = self.call_FUT('question_order.ini')
        self.assertEqual(c['questions_order'], ['foo'])

    @mock.patch('mrbob.parsing.urlretrieve')
    def test_parse_remote_config(self, urlretrieve):

        def write(url, filename):
            f = open(filename, 'w')
            f.write(
                "[variables]\n"
                "foo = bar\n"
            )
            f.close()

        urlretrieve.side_effect = write

        c = self.call_FUT(configname='http://nohost/mrbob.ini')
        self.assertEqual(c['variables']['foo'], 'bar')


class update_configTest(unittest.TestCase):

    def call_FUT(self, config, newconfig):
        from ..parsing import update_config
        return update_config(config, newconfig)

    def test_update_config_override_one_option(self):
        config = {
            'foo': 'bar',
            'foo1': 'mar'
        }
        new_config = {
            'foo1': 'bar'
        }
        self.call_FUT(config, new_config)

        expected_config = {
            'foo': 'bar',
            'foo1': 'bar'
        }

        self.assertEqual(config, expected_config)

    def test_update_config_override_nested(self):
        config = {
            'foo': 'bar',
            'bar': {
                'foo': 'bar',
                'foo1': 'foo',
            }
        }
        new_config = {
            'foo1': 'bar',
            'bar': {
                'foo1': 'moo',
                'moo': 'moo',
            }
        }
        self.call_FUT(config, new_config)

        expected_config = {
            'foo': 'bar',
            'foo1': 'bar',
            'bar': {
                'foo': 'bar',
                'foo1': 'moo',
                'moo': 'moo',
            }
        }

        self.assertEqual(config, expected_config)


class write_configTest(unittest.TestCase):

    def setUp(self):
        self.tmpfile = tempfile.mkstemp()[1]

    def tearDown(self):
        os.remove(self.tmpfile)

    def call_FUT(self, section, data):
        from ..parsing import write_config
        return write_config(self.tmpfile, section, data)

    def test_multiple_options(self):
        self.call_FUT(
            'variables',
            {'foo.bar': 'a',
             'foo.bar.moo': 'b'},
        )

        with open(self.tmpfile) as f:
            output = f.read()
            self.assertTrue('\nfoo.bar = a\n' in output)
            self.assertTrue('\nfoo.bar.moo = b\n' in output)

    def test_empty(self):
        self.call_FUT(
            'variables',
            {},
        )

        with open(self.tmpfile) as f:
            self.assertEqual(f.read(), """[variables]\n\n""")

    def test_utf8(self):
        if six.PY3:  # pragma: nocover
            var_ = 'č'
        else:  # pragma: nocover
            var_ = 'č'.decode('utf-8')

        self.call_FUT(
            'variables',
            {'foo.bar': var_},
        )

        with codecs.open(self.tmpfile, 'r', 'utf-8') as f:
            self.assertEqual(f.read(), six.u("[variables]\nfoo.bar = %s\n\n") % var_)

    def test_non_str(self):
        self.call_FUT(
            'variables',
            {'foo': True},
        )

        with open(self.tmpfile) as f:
            output = f.read()
            self.assertTrue('\nfoo = True\n' in output)


class pretty_format_configTest(unittest.TestCase):

    def call_FUT(self, config):
        from ..parsing import pretty_format_config
        return pretty_format_config(config)

    def test_complex(self):
        c = self.call_FUT({
            'foo': 'bar',
            'bar': {'moo': '1', 'ma': '2'},
            'z': 'z',
            'a': 'b',
        })
        self.assertEqual(c, [
            'a = b',
            'bar.ma = 2',
            'bar.moo = 1',
            'foo = bar',
            'z = z'],
        )


class nest_variablesTest(unittest.TestCase):

    def call_FUT(self, d):
        from ..parsing import nest_variables
        return nest_variables(d)

    def test_overwrite_dict_with_value(self):
        """ providing a value for a key that already contains a
        dictionary raises a ConfigurationError """
        from ..bobexceptions import ConfigurationError
        d = OrderedDict([('foo.bar', '1'), ('foo', '2')])
        self.assertRaises(ConfigurationError, self.call_FUT, d)

    def test_overwrite_value_with_dict(self):
        """ providing a dict for a key that already contains a
        string raises a ConfigurationError """
        from ..bobexceptions import ConfigurationError
        d = OrderedDict([('foo', '2'), ('foo.bar', '1')])
        self.assertRaises(ConfigurationError, self.call_FUT, d)



"""Command line interface to mr.bob"""

import pkg_resources
import sys
import os
import shutil

import six
import argparse

from .configurator import Configurator
from .configurator import maybe_bool
from .bobexceptions import ConfigurationError
from .bobexceptions import TemplateConfigurationError
from .parsing import parse_config, update_config, pretty_format_config


# http://docs.python.org/library/argparse.html
parser = argparse.ArgumentParser(description='Filesystem template renderer')
parser.add_argument('template',
                    nargs="?",
                    help="""Template name to use for rendering. See
                    http://mrbob.readthedocs.org/en/latest/userguide.html#usage
                    for a guide to template syntax

                    """)
parser.add_argument('-O', '--target-directory',
                    default=".",
                    dest="target_directory",
                    help='Where to output rendered structure. Defaults to current directory')
parser.add_argument('-v', '--verbose',
                    action="store_true",
                    default=False,
                    help='Print more output for debugging')
parser.add_argument('-c', '--config',
                    action="store",
                    help='Configuration file to specify either [mr.bob] or [variables] sections')
parser.add_argument('-V', '--version',
                    action="store_true",
                    default=False,
                    help='Display version number')
parser.add_argument('-l', '--list-questions',
                    action="store_true",
                    default=False,
                    help='List all questions needed for the template')
parser.add_argument('-w', '--remember-answers',
                    action="store_true",
                    default=False,
                    help='Remember answers to .mrbob.ini file inside output directory')
parser.add_argument('-n', '--non-interactive',
                  dest='non_interactive',
                  action='store_true',
                  default=False,
                  help="Don't prompt for input. Fail if questions are required but not answered")
parser.add_argument('-q', '--quiet',
                    action="store_true",
                    default=False,
                    help='Suppress all but necessary output')


def main(args=sys.argv[1:]):
    """Main function called by `mrbob` command.
    """
    options = parser.parse_args(args=args)

    if options.version:
        version = pkg_resources.get_distribution('mr.bob').version
        return version

    if not options.template:
        parser.error('You must specify what template to use.')

    userconfig = os.path.expanduser('~/.mrbob')
    if os.path.exists(userconfig):
        global_config = parse_config(userconfig)
        global_bobconfig = global_config['mr.bob']
        global_variables = global_config['variables']
        global_defaults = global_config['defaults']
    else:
        global_bobconfig = {}
        global_variables = {}
        global_defaults = {}

    original_global_bobconfig = dict(global_bobconfig)
    original_global_variables = dict(global_variables)
    original_global_defaults = dict(global_defaults)

    if options.config:
        try:
            file_config = parse_config(options.config)
        except ConfigurationError as e:
            parser.error(e)
        file_bobconfig = file_config['mr.bob']
        file_variables = file_config['variables']
        file_defaults = file_config['defaults']

    else:
        file_bobconfig = {}
        file_variables = {}
        file_defaults = {}

    cli_variables = {}  # TODO: implement variables on cli
    cli_defaults = {}  # TODO: implement defaults on cli
    cli_bobconfig = {
        'verbose': options.verbose,
        'quiet': options.quiet,
        'remember_answers': options.remember_answers,
        'non_interactive': options.non_interactive,
    }

    bobconfig = update_config(update_config(global_bobconfig, file_bobconfig), cli_bobconfig)
    variables = update_config(update_config(global_variables, file_variables), cli_variables)
    defaults = update_config(update_config(global_defaults, file_defaults), cli_defaults)

    c = None
    if bobconfig['verbose']:
        print('')
        print('Configuration provided:')
        print('')
        print('[variables] from ~/.mrbob')
        for line in pretty_format_config(original_global_variables):
            print(line)
        print('[variables] from --config file')
        for line in pretty_format_config(file_variables):
            print(line)
        # TODO: implement variables on cli
        # print('[variables] from command line interface')
        # for line in pretty_format_config(file_variables):
        #     print(line)
        print('[defaults] from ~/.mrbob')
        for line in pretty_format_config(original_global_defaults):
            print(line)
        print('[defaults] from --config file')
        for line in pretty_format_config(file_defaults):
            print(line)
        # TODO: implement defaults on cli
        # print('[defaults] from command line interface')
        # for line in pretty_format_config(file_defaults):
        #     print(line)
        print('[mr.bob] from ~/.mrbob')
        for line in pretty_format_config(original_global_bobconfig):
            print(line)
        print('[mr.bob] from --config file')
        for line in pretty_format_config(file_bobconfig):
            print(line)
        print('[mr.bob] from command line interface')
        for line in pretty_format_config(cli_bobconfig):
            print(line)

    try:
        c = Configurator(template=options.template,
            target_directory=options.target_directory,
            bobconfig=bobconfig,
            variables=variables,
            defaults=defaults)

        if options.list_questions:
            return c.print_questions()

        if c.questions and not maybe_bool(bobconfig['quiet']):
            if options.non_interactive:
                print('')
                print('Welcome to mr.bob non-interactive mode. Questions will be answered by default values or hooks.')
                print('')
            else:
                print('')
                print('Welcome to mr.bob interactive mode. Before we generate directory structure, some questions need to be answered.')
                print('')
                print('Answer with a question mark to display help.')
                print('Values in square brackets at the end of the questions show the default value if there is no answer.')
                print('\n')
            c.ask_questions()
            if not options.non_interactive:
                print('')
        c.render()
        if not maybe_bool(bobconfig['quiet']):
            print("Generated file structure at %s" % os.path.realpath(options.target_directory))
            print('')
        return
    except TemplateConfigurationError as e:
        parser.error(six.u('TemplateConfigurationError: %s') % e.args[0])
    except ConfigurationError as e:
        parser.error(six.u('ConfigurationError: %s') % e.args[0])
    finally:
        if c and c.is_tempdir:
            shutil.rmtree(c.template_dir)


if __name__ == '__main__':  # pragma: nocover
    print(main())



from os import path
from shutil import copy2
import codecs
import fnmatch
import os
import re
import six
import stat

from jinja2 import Environment, StrictUndefined


jinja2_env = Environment(
    block_start_string="{{%",
    block_end_string="%}}",
    variable_start_string="{{{",
    variable_end_string="}}}",
    trim_blocks=True,
    undefined=StrictUndefined,
)


DEFAULT_IGNORED_FILES = ['.mrbob.ini', '.DS_Store']
DEFAULT_IGNORED_DIRECTORIES = []


def jinja2_renderer(s, v):
    return jinja2_env.from_string(s).render(parse_variables(v))


def python_formatting_renderer(s, v):
    return s % v


def parse_variables(variables):
    d = dict()

    for key, value in variables.items():
        keys = key.split('.')
        new_d = None
        for k in keys[:-1]:
            if new_d is None:
                if k not in d:
                    d[k] = dict()
                new_d = d[k]
            else:
                if k not in new_d:
                    new_d[k] = dict()
                new_d = new_d[k]
        if new_d is None:
            d[keys[-1]] = value
        else:
            new_d[keys[-1]] = value
    return dict(d)


def matches_any(filename, patterns):
    result = any(fnmatch.fnmatch(filename, pat) for pat in patterns)
    return result


def render_structure(fs_source_root, fs_target_root, variables, verbose,
        renderer, ignored_files, ignored_directories):
    """Recursively copies the given filesystem path `fs_source_root_ to a target directory `fs_target_root`.

    Any files ending in `.bob` are rendered as templates using the given
    renderer using the variables dictionary, thereby losing the `.bob` suffix.

    strings wrapped in `+` signs in file- or directory names will be replaced
    with values from the variables, i.e. a file named `+name+.py.bob` given a
    dictionary {'name': 'bar'} would be rendered as `bar.py`.
    """
    ignored_files.extend(DEFAULT_IGNORED_FILES)
    ignored_directories.extend(DEFAULT_IGNORED_DIRECTORIES)
    if not isinstance(fs_source_root, six.text_type):  # pragma: no cover
        fs_source_root = six.u(fs_source_root)
    for fs_source_dir, local_directories, local_files in os.walk(fs_source_root, topdown=True):
        fs_target_dir = path.abspath(path.join(fs_target_root, path.relpath(fs_source_dir, fs_source_root)))
        local_directories[:] = [d for d in local_directories if not matches_any(d, ignored_directories)]
        for local_file in local_files:
            if matches_any(local_file, ignored_files):
                continue
            render_template(
                path.join(fs_source_dir, local_file),
                render_filename(fs_target_dir, variables),
                variables,
                verbose,
                renderer,
            )
        for local_directory in local_directories:
            abs_dir = render_filename(path.join(fs_target_dir, local_directory), variables)
            if not path.exists(abs_dir):
                if verbose:
                    print(six.u("mkdir %s") % abs_dir)
                os.mkdir(abs_dir)


def render_template(fs_source, fs_target_dir, variables, verbose, renderer):
    filename = path.split(fs_source)[1]
    if filename.endswith('.bob'):
        filename = filename.split('.bob')[0]
        fs_target_path = path.join(fs_target_dir, render_filename(filename, variables))
        if verbose:
            print(six.u("Rendering %s to %s") % (fs_source, fs_target_path))
        fs_source_mode = stat.S_IMODE(os.stat(fs_source).st_mode)
        with codecs.open(fs_source, 'r', 'utf-8') as f:
            source_output = f.read()
            output = renderer(source_output, variables)
            # append newline due to jinja2 bug, see https://github.com/iElectric/mr.bob/issues/30
            if source_output.endswith('\n') and not output.endswith('\n'):
                output += '\n'
        with codecs.open(fs_target_path, 'w', 'utf-8') as fs_target:
            fs_target.write(output)
        os.chmod(fs_target_path, fs_source_mode)
    else:
        fs_target_path = path.join(fs_target_dir, render_filename(filename, variables))
        if verbose:
            print(six.u("Copying %s to %s") % (fs_source, fs_target_path))
        copy2(fs_source, fs_target_path)
    return path.join(fs_target_dir, filename)


def render_filename(filename, variables):
    variables_regex = re.compile(r"\+[^+%s]+\+" % re.escape(os.sep))

    replaceables = variables_regex.findall(filename)
    for replaceable in replaceables:
        actual_replaceable = replaceable.replace('+', '')
        if actual_replaceable in variables:
            filename = filename.replace(replaceable, variables[actual_replaceable])
        else:
            raise KeyError('%s key part of filename %s was not found in variables %s' % (actual_replaceable, filename, variables))
    return filename



"""Use any of hooks below or write your own. You are welcome to contribute them!"""

import re
import time

from .bobexceptions import ValidationError


def to_boolean(configurator, question, answer):
    """
    If you want to convert an answer to Python boolean, you can
    use this function as :ref:`post-question-hook`:

    .. code-block:: ini

        [questions]
        idiot.question = Are you young?
        idiot.post_ask_question = mrbob.hooks:to_boolean

    Following variables can be converted to a boolean: **y, n, yes, no, true, false, 1, 0**
    """
    value = answer.lower()
    if value in ['y', 'yes', 'true', '1']:
        return True
    elif value in ['n', 'no', 'false', '0']:
        return False
    else:
        raise ValidationError('Value must be a boolean (y/n)')


def to_integer(configurator, question, answer):
    """
    If you want to convert an answer to Python integer, you can
    use this function as :ref:`post-question-hook`:

    .. code-block:: ini

        [questions]
        owner.question = What's your age?
        owner.post_ask_question = mrbob.hooks:to_integer

    """
    try:
        return int(answer)
    except ValueError:
        raise ValidationError('Value must be an integer')


def to_decimal(configurator, question, answer):
    """
    If you want to convert an answer to Python float, you can
    use this function as :ref:`post-question-hook`:

    .. code-block:: ini

        [questions]
        daemon.question = What interval should the daemon poll for data?
        daemon.post_ask_question = mrbob.hooks:to_decimal

    """
    try:
        return float(answer)
    except ValueError:
        raise ValidationError('Value must be a decimal')


def validate_choices(configurator, question, answer):
    """
    If you want to validate an answer using a limited set of choices, you can
    use this function as :ref:`post-question-hook`:

    .. code-block:: ini

        [questions]
        license.question = What license would you like to use?
        license.post_ask_question = mrbob.hooks:validate_choices
        license.choices = MIT|BSD|Apache 2.0|Other
        license.choices_case_sensitive = yes
        license.choices_delimiter = |

    Currently choices are split using whitespace by default. If you wish to
    have whitespace within each choice, you may specify a custom delimiter
    which will be used to split the choices.

    This hook may be set to validate the choices in a case sensitive manner.
    However, this behaviour is disabled by default.
    """
    delimiter = question.extra.get('choices_delimiter')
    choices = question.extra.get('choices', '').split(delimiter)

    # If no choices are defined, then we assume the provided answer is correct
    if not choices:
        return answer

    # Determine case sensitivity
    case_sensitive_config = question.extra.get('choices_case_sensitive')
    case_sensitive = False
    if case_sensitive_config:
        try:
            case_sensitive = to_boolean(None, None, case_sensitive_config)
        except ValidationError:
            pass

    if case_sensitive:
        valid = answer in choices
    else:
        valid = answer.lower() in [c.lower() for c in choices]

    if valid:
        return answer
    else:
        raise ValidationError(
            'Value must be ' + ', '.join(choices[:-1]) + ' or ' + choices[-1])


def validate_regex(configurator, question, answer):
    """
    If you want to validate an answer using a regular expression, you can
    use this function as :ref:`post-question-hook`:

    .. code-block:: ini

        [questions]
        project.question = Please specify a name (only lowercase characters)?
        project.post_ask_question = mrbob.hooks:validate_regex
        project.regex = ^[a-z]+$

    """
    regex = question.extra.get('regex')

    # If no regex is defined, then we assume the provided answer is correct
    if not regex:
        return answer

    if re.match(regex, answer):
        return answer
    else:
        raise ValidationError(
            'Value was not of the expected format (%s)' % regex)


def set_current_datetime(configurator, question):
    """
    If you want to set the default answer of a question to the current
    date and time, use this function as :ref:`pre-question-hook`:

    .. code-block:: ini

        [questions]
        project.question = What year was the project started?
        project.pre_ask_question = mrbob.hooks:set_current_datetime
        project.datetime_format = %%Y

    The datetime_format property should be of the standard Python strftime
    format.  It defaults to YYYY-MM-DD if not specified.

    Please note that you'll have to escape the % character (by using %%)
    due to the fact it's a special character in INI files.

    See the following URL for more information:
    http://docs.python.org/2/library/datetime.html#strftime-and-strptime-behavior
    """
    datetime_format = question.extra.get('datetime_format', '%Y-%m-%d')
    question.default = time.strftime(datetime_format)


def validate_datetime(configurator, question, answer):
    """
    If you want to validate a date using a chosen date format, you can
    use this function as :ref:`post-question-hook`:

    .. code-block:: ini

        [questions]
        project.question = What year was the project started?
        project.post_ask_question = mrbob.hooks:validate_datetime
        project.datetime_format = %%Y

    The datetime_format property should be of the standard Python strftime
    format.  It defaults to YYYY-MM-DD if not specified.

    Please note that you'll have to escape the % character (by using %%)
    due to the fact it's a special character in INI files.

    See the following URL for more information:
    http://docs.python.org/2/library/datetime.html#strftime-and-strptime-behavior
    """
    datetime_format = question.extra.get('datetime_format', '%Y-%m-%d')
    try:
        time.strptime(answer, datetime_format)
        return answer
    except ValueError:
        raise ValidationError(
            'Value was not a date in the format ' + datetime_format)


def show_message(configurator):
    """
    If you want to display a message to the user when rendering is complete, you
    can use this function as :ref:`post-render-hook`:

    .. code-block:: ini

        [template]
        post_render = mrbob.hooks:show_message
        message = Well done, %%(author.name)s, your code is ready!

    As shown above, you can use standard Python formatting in ``post_render_msg``.
    """
    message = configurator.templateconfig.get('message', '')
    if not configurator.quiet and message:
        print(message % configurator.variables)



# -*- coding: utf-8 -*-

import unittest
import os
import sys
import tempfile
import shutil
import six
from unittest import mock


mocked_pre_ask_question = mock.Mock()
mocked_post_ask_question = mock.Mock()
mocked_post_ask_question_validationerror = mock.Mock()
mocked_post_ask_question_validationerror_non_interactive = mock.Mock()
mocked_render_hook = mock.Mock()
mocked_ask_hook = mock.Mock()


def dummy_prompt(value):  # pragma: no cover
    pass


def dummy_renderer(value):  # pragma: no cover
    pass


def dummy_question_hook(configurator, question):  # pragma: no cover
    return


def dummy_question_hook2(configurator, question):  # pragma: no cover
    return


def dummy_render_hook(configurator):  # pragma: no cover
    return


def dummy_question_hook_skipquestion(configurator, question):  # pragma: no cover
    from ..bobexceptions import SkipQuestion

    raise SkipQuestion


class DummyConfigurator(object):
    def __init__(
        self,
        defaults=None,
        bobconfig=None,
        templateconfig=None,
        variables=None,
        quiet=False,
    ):
        self.defaults = defaults or {}
        self.bobconfig = bobconfig or {}
        self.variables = variables or {}
        self.quiet = quiet
        self.templateconfig = templateconfig or {}


class resolve_dotted_pathTest(unittest.TestCase):
    def call_FUT(self, name):
        from ..configurator import resolve_dotted_path

        return resolve_dotted_path(name)

    def test_nomodule(self):
        self.assertRaises(ImportError, self.call_FUT, "foobar.blabla:foo")

    def test_return_abs_path(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates")
        abs_path = self.call_FUT("mrbob.tests:templates")
        self.assertEqual(abs_path, template_dir)


class resolve_dotted_funcTest(unittest.TestCase):
    def call_FUT(self, name):
        from ..configurator import resolve_dotted_func

        return resolve_dotted_func(name)

    def test_nomodule(self):
        self.assertRaises(ImportError, self.call_FUT, "foobar.blabla:foo")

    def test_error_no_func(self):
        from ..bobexceptions import ConfigurationError

        self.assertRaises(ConfigurationError, self.call_FUT, "mrbob.rendering:foo")

    def test_return_func(self):
        from mrbob.rendering import jinja2_renderer

        func = self.call_FUT("mrbob.rendering:jinja2_renderer")
        self.assertEqual(func, jinja2_renderer)


class parse_templateTest(unittest.TestCase):
    def call_FUT(self, name):
        from ..configurator import parse_template

        return parse_template(name)

    def test_relative(self):
        old_cwd = os.getcwd()
        os.chdir(os.path.dirname(__file__))
        template_dir = os.path.join(os.path.dirname(__file__), "templates")
        abs_path = self.call_FUT("templates")
        os.chdir(old_cwd)
        self.assertEqual(abs_path, (template_dir, False))

    def test_absolute(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates")
        abs_path = self.call_FUT(template_dir)
        self.assertEqual(abs_path, (template_dir, False))

    def test_dotted(self):
        template_dir = os.path.join(os.path.dirname(__file__), "templates")
        abs_path = self.call_FUT("mrbob.tests:templates")
        self.assertEqual(abs_path, (template_dir, False))

    def test_not_a_dir(self):
        from ..bobexceptions import ConfigurationError

        self.assertRaises(ConfigurationError, self.call_FUT, "foo_bar")

    @mock.patch("mrbob.configurator.urlretrieve")
    def test_zipfile(self, mock_urlretrieve):
        mock_urlretrieve.side_effect = self.fake_zip
        abs_path = self.call_FUT("http://foobar.com/bla.zip")
        self.assertEqual(set(os.listdir(abs_path[0])), set(["test", ".mrbob.ini"]))

    @mock.patch("mrbob.configurator.urlretrieve")
    def test_zipfile_base_path(self, mock_urlretrieve):
        mock_urlretrieve.side_effect = self.fake_zip_base_path
        abs_path = self.call_FUT("http://foobar.com/bla.zip#some/dir")
        self.assertEqual(set(os.listdir(abs_path[0])), set(["test", ".mrbob.ini"]))

    @mock.patch("mrbob.configurator.urlretrieve")
    def test_zipfile_not_zipfile(self, mock_urlretrieve):
        from ..bobexceptions import ConfigurationError

        mock_urlretrieve.side_effect = self.fake_wrong_zip
        self.assertRaises(
            ConfigurationError, self.call_FUT, "http://foobar.com/bla.tar#some/dir"
        )

    def fake_wrong_zip(self, url, path):
        with open(path, "w") as f:
            f.write("boo")

    def fake_zip(self, url, path):
        import zipfile

        zf = zipfile.ZipFile(path, "w")
        try:
            zf.writestr(".mrbob.ini", "[questions]\n")
            zf.writestr("test", "test")
        finally:
            zf.close()

    def fake_zip_base_path(self, url, path):
        import zipfile

        zf = zipfile.ZipFile(path, "w")
        try:
            zf.writestr("some/dir/.mrbob.ini", "[questions]\n")
            zf.writestr("some/dir/test", "test")
        finally:
            zf.close()


class ConfiguratorTest(unittest.TestCase):
    def setUp(self):
        self.target_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.target_dir)

    def call_FUT(self, *args, **kw):
        from ..configurator import Configurator

        return Configurator(*args, **kw)

    def test_target_directory_inside_template_dir(self):
        from ..bobexceptions import ConfigurationError

        self.assertRaises(
            ConfigurationError,
            self.call_FUT,
            "mrbob.tests:templates/questions1",
            os.path.join(os.path.dirname(__file__), "templates/questions1/foo"),
            {},
        )

    def test_parse_questions_basic(self):
        c = self.call_FUT("mrbob.tests:templates/questions1", self.target_dir, {})
        self.assertEqual(len(c.questions), 2)
        self.assertEqual(c.questions[0].name, "foo.bar.car.dar")
        self.assertEqual(c.questions[0].question, "Why?")
        self.assertEqual(c.questions[1].name, "foo")
        self.assertEqual(c.questions[1].question, "What?")

    def test_parse_questions_no_questions(self):
        c = self.call_FUT("mrbob.tests:templates/questions2", self.target_dir, {})
        self.assertEqual(len(c.questions), 0)

    def test_parse_questions_no_questions_section(self):
        self.call_FUT("mrbob.tests:templates/empty2", self.target_dir, {})

    def test_parse_questions_extra_parameter(self):
        c = self.call_FUT("mrbob.tests:templates/questions3", self.target_dir, {})
        self.assertEqual(c.questions[0].extra, {"foobar": "something"})

    def test_parse_questions_all(self):
        c = self.call_FUT("mrbob.tests:templates/questions4", self.target_dir, {})
        self.assertEqual(len(c.questions), 1)
        self.assertEqual(c.questions[0].name, six.u("foo"))
        self.assertEqual(c.questions[0].default, "True")
        self.assertEqual(c.questions[0].required, False)
        self.assertEqual(c.questions[0].help, six.u("Blabla blabal balasd a a sd"))
        self.assertEqual(c.questions[0].command_prompt, dummy_prompt)

    def test_ask_questions_empty(self):
        args = ["mrbob.tests:templates/questions1", self.target_dir, {}]
        c = self.call_FUT(*args)
        c.questions = []
        c.variables = {}
        c.ask_questions()
        self.assertEqual(c.variables, {})

    def test_pre_post_ask_hooks_multiple(self):
        c = self.call_FUT(
            "mrbob.tests:templates/ask_hooks",
            self.target_dir,
            {},
        )
        self.assertEqual(c.pre_ask, [dummy_render_hook, mocked_ask_hook])
        self.assertEqual(c.post_ask, [dummy_render_hook, mocked_ask_hook])
        c.ask_questions()
        self.assertEqual(mocked_ask_hook.mock_calls, [mock.call(c), mock.call(c)])

    def test_ask_questions_missing(self):
        from ..configurator import Question

        args = ["mrbob.tests:templates/questions1", self.target_dir, {}]
        c = self.call_FUT(*args)
        c.questions = [
            Question("foo.bar", "fobar?"),
            Question("moo", "Moo?", command_prompt=lambda x: "moo."),
        ]
        c.variables = {"foo.bar": "answer"}
        c.ask_questions()
        self.assertEqual(c.variables, {"foo.bar": "answer", "moo": "moo."})

    @mock.patch("mrbob.configurator.render_structure")
    def test_remember_answers(self, mock_render_structure):
        args = [
            "mrbob.tests:templates/questions1",
            self.target_dir,
            {"remember_answers": "True"},
            {"foo.bar": "3"},
        ]
        c = self.call_FUT(*args)
        c.render()
        with open(os.path.join(self.target_dir, ".mrbob.ini")) as f:
            self.assertEqual(f.read().strip(), """[variables]\nfoo.bar = 3""".strip())

    @mock.patch("mrbob.configurator.render_structure")
    def test_remember_answers_default(self, mock_render_structure):
        c = self.call_FUT(
            "mrbob.tests:templates/questions1",
            self.target_dir,
            variables={"foo.bar": "3"},
        )
        c.render()
        self.assertFalse(os.path.exists(os.path.join(self.target_dir, ".mrbob.ini")))

    def test_renderer_default(self):
        from ..rendering import jinja2_renderer

        c = self.call_FUT("mrbob.tests:templates/empty", self.target_dir, {})
        self.assertEqual(c.renderer, jinja2_renderer)

    def test_renderer_set(self):
        c = self.call_FUT("mrbob.tests:templates/renderer", self.target_dir, {})
        self.assertEqual(c.renderer, dummy_renderer)

    def test_pre_post_render_hooks_multiple(self):
        c = self.call_FUT(
            "mrbob.tests:templates/render_hooks",
            self.target_dir,
            {},
        )
        self.assertEqual(c.pre_render, [dummy_render_hook, mocked_render_hook])
        self.assertEqual(c.post_render, [dummy_render_hook, mocked_render_hook])
        c.render()
        self.assertEqual(mocked_render_hook.mock_calls, [mock.call(c), mock.call(c)])

    def test_ignored_files(self):
        c = self.call_FUT("mrbob.tests:templates/ignored", self.target_dir, {})
        self.assertEqual(len(c.ignored_files), 2)
        self.assertTrue("ignored" in c.ignored_files)
        self.assertTrue("*.txt" in c.ignored_files)

    def test_ignored_directories(self):
        c = self.call_FUT("mrbob.tests:templates/ignored_dirs", self.target_dir, {})
        self.assertEqual(len(c.ignored_directories), 2)
        self.assertTrue("ignored" in c.ignored_directories)
        self.assertTrue("*_stuff" in c.ignored_directories)


class QuestionTest(unittest.TestCase):
    def call_FUT(self, *args, **kw):
        from ..configurator import Question

        return Question(*args, **kw)

    def test_defaults(self):
        from six import moves

        q = self.call_FUT("foo", "Why?")
        self.assertEqual(q.name, "foo")
        self.assertEqual(q.default, None)
        self.assertEqual(q.required, False)
        self.assertEqual(q.help, "")
        self.assertEqual(q.command_prompt, moves.input)

    def test_repr(self):
        q = self.call_FUT("foo", "Why?")
        self.assertEqual(
            repr(q),
            six.u("<Question name=foo question='Why?' default=None required=False>"),
        )

    def test_ask(self):
        def cmd(q):
            self.assertEqual(q, "--> Why?: ")
            return "foo"

        q = self.call_FUT("foo", "Why?", command_prompt=cmd)
        answer = q.ask(DummyConfigurator())
        self.assertEqual(answer, "foo")

    def test_ask_unicode(self):
        def cmd(q):
            self.assertTrue(isinstance(q, str))
            return "foo"

        q = self.call_FUT("foo", six.u("č?"), command_prompt=cmd)
        q.ask(DummyConfigurator())

    def test_ask_default_empty(self):
        q = self.call_FUT("foo", "Why?", default="moo", command_prompt=lambda x: "")
        answer = q.ask(DummyConfigurator())
        self.assertEqual(answer, "moo")

    def test_ask_default_not_empty(self):
        def cmd(q):
            self.assertEqual(q, "--> Why? [moo]: ")
            return "foo"

        q = self.call_FUT("foo", "Why?", default="moo", command_prompt=cmd)
        answer = q.ask(DummyConfigurator())
        self.assertEqual(answer, "foo")

    def test_ask_no_default_and_not_required(self):
        def cmd(q, go=["foo", ""]):
            return go.pop()

        q = self.call_FUT("foo", "Why?", command_prompt=cmd)
        answer = q.ask(DummyConfigurator())
        self.assertEqual(answer, "")

    def test_ask_no_default_and_required(self):
        def cmd(q, go=["foo", ""]):
            return go.pop()

        q = self.call_FUT("foo", "Why?", required=True, command_prompt=cmd)
        answer = q.ask(DummyConfigurator())
        self.assertEqual(answer, "foo")

    def test_ask_no_help(self):
        from six import StringIO

        def cmd(q, go=["foo", "?"]):
            return go.pop()

        sys.stdout = StringIO()
        q = self.call_FUT("foo", "Why?", command_prompt=cmd)
        q.ask(DummyConfigurator())
        self.assertEqual(sys.stdout.getvalue(), "There is no additional help text.\n\n")
        sys.stdout = sys.__stdout__

    def test_ask_help(self):
        from six import StringIO

        def cmd(q, go=["foo", "?"]):
            return go.pop()

        sys.stdout = StringIO()
        q = self.call_FUT("foo", "Why?", help="foobar_help", command_prompt=cmd)
        q.ask(DummyConfigurator())
        self.assertEqual(sys.stdout.getvalue(), "foobar_help\n\n")
        sys.stdout = sys.__stdout__

    def test_non_interactive_required(self):
        from ..bobexceptions import ConfigurationError

        q = self.call_FUT("foo", "Why?", required=True)
        c = DummyConfigurator(bobconfig={"non_interactive": "True"})
        self.assertRaises(ConfigurationError, q.ask, c)

    def test_non_interactive_not_required(self):
        q = self.call_FUT("foo", "Why?")
        c = DummyConfigurator(bobconfig={"non_interactive": "True"})
        answer = q.ask(c)
        self.assertEqual(answer, "")

    def test_defaults_override(self):
        q = self.call_FUT("foo", "Why?", default="foo")
        c = DummyConfigurator(
            bobconfig={"non_interactive": "True"}, defaults={"foo": "moo"}
        )
        answer = q.ask(c)
        self.assertEqual(answer, "moo")

    def test_pre_ask_question(self):
        q = self.call_FUT(
            "foo",
            "Why?",
            command_prompt=lambda x: "",
            pre_ask_question="mrbob.tests.test_configurator:mocked_pre_ask_question",
        )
        c = DummyConfigurator()
        q.ask(c)
        mocked_pre_ask_question.assert_called_with(c, q)

    def test_pre_ask_question_multiple(self):
        q = self.call_FUT(
            "foo",
            "Why?",
            pre_ask_question="mrbob.tests.test_configurator:dummy_question_hook mrbob.tests.test_configurator:dummy_question_hook2",
        )
        self.assertEqual(
            q.pre_ask_question, [dummy_question_hook, dummy_question_hook2]
        )

    def test_pre_ask_question_skipquestion(self):
        q = self.call_FUT(
            "foo",
            "Why?",
            pre_ask_question="mrbob.tests.test_configurator:dummy_question_hook_skipquestion",
        )
        self.assertEqual(q.ask(DummyConfigurator()), None)

    def test_post_ask_question(self):
        q = self.call_FUT(
            "foo",
            "Why?",
            command_prompt=lambda x: "",
            post_ask_question="mrbob.tests.test_configurator:mocked_post_ask_question",
        )
        c = DummyConfigurator()
        answer = q.ask(c)
        mocked_post_ask_question.assert_called_with(c, q, "")
        self.assertEqual(mocked_post_ask_question(), answer)

    def test_post_ask_question_multiple(self):
        q = self.call_FUT(
            "foo",
            "Why?",
            post_ask_question="mrbob.tests.test_configurator:dummy_question_hook mrbob.tests.test_configurator:dummy_question_hook2",
        )
        self.assertEqual(
            q.post_ask_question, [dummy_question_hook, dummy_question_hook2]
        )

    def test_post_ask_question_validationerror(self):
        def cmd(q, go=["bar", "foo"]):
            return go.pop()

        def side_effect(configurator, question, answer):
            from ..bobexceptions import ValidationError

            if answer == "foo":
                raise ValidationError
            elif answer == "bar":
                return "moo"

        mocked_post_ask_question_validationerror.side_effect = side_effect

        q = self.call_FUT(
            "foo",
            "Why?",
            command_prompt=cmd,
            post_ask_question="mrbob.tests.test_configurator:mocked_post_ask_question_validationerror",
        )
        c = DummyConfigurator()
        self.assertEqual(q.ask(c), "moo")

    def test_post_ask_question_validationerror_non_interactive(self):
        from ..bobexceptions import ConfigurationError, ValidationError

        mocked_post_ask_question_validationerror_non_interactive.side_effect = (
            ValidationError
        )

        q = self.call_FUT(
            "foo",
            "Why?",
            command_prompt=lambda x: "",
            post_ask_question="mrbob.tests.test_configurator:mocked_post_ask_question_validationerror_non_interactive",
        )
        c = DummyConfigurator(bobconfig={"non_interactive": "True"})
        self.assertRaises(ConfigurationError, q.ask, c)



"""mr.bob exceptions module."""


class MrBobError(Exception):
    """Base class for errors"""


class ConfigurationError(MrBobError):
    """Raised during configuration phase"""


class TemplateConfigurationError(ConfigurationError):
    """Raised reading template configuration"""


class ValidationError(MrBobError):
    """Raised during question validation"""


class SkipQuestion(MrBobError):
    """Raised during pre_ask_question if we should skip it"""



import os
import tempfile
import collections

try:  # pragma: no cover
    from collections import OrderedDict  # NOQA
except ImportError:  # pragma: no cover
    from ordereddict import OrderedDict  # NOQA
import six
import configparser

try:  # pragma: no cover
    from urllib import urlretrieve  # NOQA
except ImportError:  # pragma: no cover
    # PY3K
    from urllib.request import urlretrieve  # NOQA

from .bobexceptions import ConfigurationError


def nest_variables(variables):
    nested = dict()
    for key, value in variables.items():
        segments = key.split(".")
        location = nested
        for segment in segments[:-1]:
            if segment not in location:
                location[segment] = dict()
            location = location[segment]
            if not isinstance(location, dict):
                raise ConfigurationError(
                    'Cannot assign "%s" to group "%s", subgroup is already used.'
                    % (value, key)
                )

        k = segments[-1]
        if isinstance(location.get(k, None), dict):
            raise ConfigurationError(
                'Cannot assign "%s" to group "%s", subgroup is already used.'
                % (value, k)
            )
        if six.PY3:  # pragma: no cover
            location[k] = value
        else:  # pragma: no cover
            location[k] = value.decode("utf-8")
    return nested


def parse_config(configname):
    tmpfile = None
    if configname.startswith("http"):
        tmpfile = tempfile.NamedTemporaryFile()
        urlretrieve(configname, tmpfile.name)
        configname = tmpfile.name

    if not os.path.exists(configname):
        raise ConfigurationError("config file does not exist: %s" % configname)

    parser = configparser.ConfigParser(dict_type=OrderedDict)
    parser.read(configname)
    config = dict()
    for section in ["variables", "defaults", "mr.bob", "questions", "template"]:
        if parser.has_section(section):
            items = parser.items(section)
            if section == "questions":
                config[section + "_order"] = [
                    key[:-9] for key, value in items if key.endswith(".question")
                ]
            if section in ["variables", "defaults"]:
                if six.PY3:  # pragma: no cover
                    config[section] = dict(items)
                else:  # pragma: no cover
                    config[section] = dict(
                        [(key, value.decode("utf-8")) for key, value in items]
                    )
            else:
                config[section] = nest_variables(dict(items))
        else:
            config[section] = {}

    if tmpfile:
        tmpfile.close()

    return config


def write_config(fs_config, section, data):
    parser = configparser.ConfigParser(dict_type=OrderedDict)
    parser.add_section(section)
    for key, value in data.items():
        if not isinstance(value, six.string_types):
            value = str(value)

        if not six.PY3:  # pragma: no cover
            value = value.encode("utf-8")
        parser.set(section, key, value)
    with open(fs_config, "w") as f:
        parser.write(f)


def update_config(first_config, second_config):
    for k, v in second_config.items():
        if isinstance(v, collections.abc.Mapping):
            r = update_config(first_config.get(k, {}), v)
            first_config[k] = r
        else:
            first_config[k] = second_config[k]
    return first_config


def pretty_format_config(config):
    l = []

    def format_config(dict_, namespace=""):
        for key, value in dict_.items():
            if namespace:
                namespace_new = namespace + ".%s" % key
            else:
                namespace_new = key
            if isinstance(value, dict):
                format_config(value, namespace=namespace_new)
            else:
                l.append("%s = %s" % (namespace_new, value))

    format_config(config)

    return sorted(l)



# -*- coding: utf-8 -*-

import os
import codecs

from setuptools import setup
from setuptools import find_packages

install_requires = [
    "setuptools",
    "six",
    "Jinja2",
]


def read(*rnames):
    return codecs.open(
        os.path.join(os.path.dirname(__file__), *rnames), "r", "utf-8"
    ).read()


setup(
    name="mr.bob",
    version="1.0.1",
    description="Bob renders directory structure templates",
    long_description=read("README.rst") + "\n" + read("HISTORY.rst"),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    keywords="template skeleton",
    author="Domen Kozar, Tom Lazar",
    author_email="",
    url="https://github.com/collective/mr.bob.git",
    license="BSD",
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=install_requires,
    extras_require={
        "test": [
            "nose2",
            "coverage",
            "flake8",
        ],
        "development": [
            "zest.releaser",
            "Sphinx",
        ],
    },
    entry_points="""
      [console_scripts]
      mrbob = mrbob.cli:main
      """,
    include_package_data=True,
    zip_safe=False,
)



from unittest import TestCase
from unittest import mock
import time

from ..configurator import Question
from .test_configurator import DummyConfigurator


class to_booleanTest(TestCase):
    def call_FUT(self, answer, configurator=None, question=None):
        from ..hooks import to_boolean

        return to_boolean(configurator or DummyConfigurator(), question, answer)

    def test_boolean(self):
        for value in ["y", "Y", "yes", "True", "1"]:
            self.assertTrue(self.call_FUT(value))
        for value in ["n", "N", "no", "False", "0"]:
            self.assertFalse(self.call_FUT(value))

    def test_boolean_wrong_input(self):
        from ..bobexceptions import ValidationError

        self.assertRaises(ValidationError, self.call_FUT, "foo")


class to_integerTest(TestCase):
    def call_FUT(self, answer, configurator=None, question=None):
        from ..hooks import to_integer

        return to_integer(configurator or DummyConfigurator(), question, answer)

    def test_integer(self):
        for value in [10, 415, -24, -1001, 0]:
            self.assertEqual(self.call_FUT(value), value)

    def test_integer_wrong_input(self):
        from ..bobexceptions import ValidationError

        self.assertRaises(ValidationError, self.call_FUT, "foo")


class to_decimalTest(TestCase):
    def call_FUT(self, answer, configurator=None, question=None):
        from ..hooks import to_decimal

        return to_decimal(configurator or DummyConfigurator(), question, answer)

    def test_decimal(self):
        for value in [10.5, 10, 415.23, -24.0, -1001.231, 0, 0.0]:
            self.assertEqual(self.call_FUT(value), value)

    def test_decimal_wrong_input(self):
        from ..bobexceptions import ValidationError

        self.assertRaises(ValidationError, self.call_FUT, "foo")


class validate_choicesTest(TestCase):
    def call_FUT(self, answer, configurator=None, question=None):
        from ..hooks import validate_choices

        return validate_choices(configurator or DummyConfigurator(), question, answer)

    def test_choices_default(self):
        q = Question(name="dummy", question="dummy", choices="MIT BSD Apache Other")
        self.assertEqual(self.call_FUT("bSd", question=q), "bSd")

    def test_choices_default_wrong_input(self):
        from ..bobexceptions import ValidationError

        q = Question(name="dummy", question="dummy", choices="MIT BSD Apache Other")
        self.assertRaises(ValidationError, self.call_FUT, "foo", None, q)

    def test_choices_case_sensitive(self):
        q = Question(
            name="dummy",
            question="dummy",
            choices="MIT BSD Apache Other",
            choices_case_sensitive="y",
        )
        self.assertEqual(self.call_FUT("BSD", question=q), "BSD")

    def test_choices_case_sensitive_wrong_input(self):
        from ..bobexceptions import ValidationError

        q = Question(
            name="dummy",
            question="dummy",
            choices="MIT BSD Apache Other",
            choices_case_sensitive="y",
        )
        self.assertRaises(ValidationError, self.call_FUT, "bsd", None, q)

    def test_choices_case_sensitive_invalid(self):
        q = Question(
            name="dummy",
            question="dummy",
            choices="MIT BSD Apache Other",
            choices_case_sensitive="allo",
        )
        self.assertEqual(self.call_FUT("bsd", question=q), "bsd")

    def test_choices_delimiter(self):
        q = Question(
            name="dummy",
            question="dummy",
            choices="MIT;BSD;Apache 2.0",
            choices_delimiter=";",
        )
        self.assertEqual(self.call_FUT("apache 2.0", question=q), "apache 2.0")

    def test_choices_delimiter_wrong_input(self):
        from ..bobexceptions import ValidationError

        q = Question(
            name="dummy",
            question="dummy",
            choices="MIT;BSD;Apache 2.0",
            choices_delimiter=";",
        )
        self.assertRaises(ValidationError, self.call_FUT, "mit;bsd", None, q)

    def test_choices_missing(self):
        q = Question(name="dummy", question="dummy")
        self.assertEqual(self.call_FUT("MIT", question=q), "MIT")


class validate_regexTest(TestCase):
    def call_FUT(self, answer, configurator=None, question=None):
        from ..hooks import validate_regex

        return validate_regex(configurator or DummyConfigurator(), question, answer)

    def test_regex(self):
        q = Question(name="dummy", question="dummy", regex="^[a-z]+[0-9]+$")
        self.assertEqual(self.call_FUT("abc123", question=q), "abc123")

    def test_regex_wrong_input(self):
        from ..bobexceptions import ValidationError

        q = Question(name="dummy", question="dummy", regex="^[a-z]+[0-9]+$")
        self.assertRaises(ValidationError, self.call_FUT, "Wr0ng", None, q)

    def test_regex_missing(self):
        q = Question(name="dummy", question="dummy")
        self.assertEqual(self.call_FUT("abc123", question=q), "abc123")


class set_current_datetimeTest(TestCase):
    def call_FUT(self, configurator=None, question=None):
        from ..hooks import set_current_datetime

        return set_current_datetime(configurator or DummyConfigurator(), question)

    def test_set_datetime_default(self):
        q = Question(name="dummy", question="dummy")
        self.call_FUT(question=q)
        self.assertEqual(q.default, time.strftime("%Y-%m-%d"))

    def test_set_datetime_custom(self):
        q = Question(name="dummy", question="dummy", datetime_format="%Y")
        self.call_FUT(question=q)
        self.assertEqual(q.default, time.strftime("%Y"))


class validate_datetimeTest(TestCase):
    def call_FUT(self, answer, configurator=None, question=None):
        from ..hooks import validate_datetime

        return validate_datetime(configurator or DummyConfigurator(), question, answer)

    def test_validate_datetime_default(self):
        q = Question(name="dummy", question="dummy")
        self.assertEqual(self.call_FUT("2014-03-01", question=q), "2014-03-01")

    def test_validate_datetime_default_wrong_input(self):
        from ..bobexceptions import ValidationError

        q = Question(name="dummy", question="dummy")
        self.assertRaises(ValidationError, self.call_FUT, "foo", None, q)

    def test_validate_datetime_custom(self):
        q = Question(name="dummy", question="dummy", datetime_format="%Y")
        self.assertEqual(self.call_FUT("2014", question=q), "2014")

    def test_validate_datetime_custom_wrong_input(self):
        from ..bobexceptions import ValidationError

        q = Question(name="dummy", question="dummy", datetime_format="%Y")
        self.assertRaises(ValidationError, self.call_FUT, "foo", None, q)


class show_messageTest(TestCase):
    def call_FUT(self, configurator=None):
        from ..hooks import show_message

        return show_message(configurator)

    @mock.patch("sys.stdout")
    def test_hook_no_msg(self, mock_stdout):
        c = DummyConfigurator()
        self.call_FUT(c)
        self.assertEqual(mock_stdout.mock_calls, [])

    @mock.patch("sys.stdout")
    def test_hook_msg_formatting(self, mock_stdout):
        c = DummyConfigurator(
            variables={"foo": "bar"}, templateconfig={"message": "Hello %(foo)s!"}
        )
        self.call_FUT(c)
        mock_stdout.write.assert_any_call("Hello bar!")

    @mock.patch("sys.stdout")
    def test_hook_quiet(self, mock_stdout):
        c = DummyConfigurator(
            variables={"foo": "bar"},
            quiet=True,
            templateconfig={"message": "Hello %(foo)s!"},
        )
        self.call_FUT(c)
        self.assertEqual(mock_stdout.mock_calls, [])
